@@ -10,9 +10,6 @@
  * Defaults
  */
 
-#define HTTP_DEFAULT_PORT 80
-#define HTTPS_DEFAULT_PORT 443
-
 #define HTTP_CLIENT_CONTINUE_TIMEOUT_MSECS (1000*2)
 #define HTTP_CLIENT_DEFAULT_REQUEST_TIMEOUT_MSECS (1000*60*1)
 #define HTTP_CLIENT_DEFAULT_DNS_LOOKUP_TIMEOUT_MSECS (1000*10)
@@ -100,7 +97,10 @@ struct http_client_request {
 	struct http_client_queue *queue;
 	struct http_client_peer *peer;
 	struct http_client_connection *conn;
+
 	struct event *event;
+	const char *const *event_headers;
+	unsigned int last_status;
 
 	string_t *headers;
 	time_t date;
@@ -109,17 +109,31 @@ struct http_client_request {
 	uoff_t payload_size, payload_offset;
 	struct ostream *payload_output;
 
+	/* Time when request can be sent the next time. This is set by
+	   http_client_request_delay*(). Default is 0 = immediately. Retries
+	   can update this. */
 	struct timeval release_time;
+	/* Time when http_client_request_submit() was called. */
 	struct timeval submit_time;
+	/* Time when the request was first sent to the server. The HTTP
+	   connection already exists at this time. */
 	struct timeval first_sent_time;
+	/* Time when the request was last sent to the server (if it was
+	   retried). */
 	struct timeval sent_time;
+	/* Time when the HTTP response header was last received. */
 	struct timeval response_time;
+	/* Time when the request will be aborted. Set by
+	   http_client_request_set_timeout(). */
 	struct timeval timeout_time;
 	unsigned int timeout_msecs;
 	unsigned int attempt_timeout_msecs;
 	unsigned int max_attempts;
 
-	unsigned int attempts;
+	uoff_t response_offset, request_offset;
+	uoff_t bytes_in, bytes_out;
+
+	unsigned int attempts, send_attempts;
 	unsigned int redirects;
 	uint64_t sent_global_ioloop_usecs;
 	uint64_t sent_http_ioloop_usecs;
@@ -166,9 +180,6 @@ struct http_client_connection {
 	struct http_client_peer_pool *ppool;
 	struct http_client_peer *peer;
 
-	char *label;
-	unsigned int id; // DEBUG: identify parallel connections
-
 	int connect_errno;
 	struct timeval connect_start_timestamp;
 	struct timeval connected_timestamp;
@@ -191,7 +202,6 @@ struct http_client_connection {
 	bool connected:1;           /* connection is connected */
 	bool tunneling:1;           /* last sent request turns this
 	                               connection into tunnel */
-	bool connect_initialized:1; /* connection was initialized */
 	bool connect_succeeded:1;   /* connection succeeded including SSL */
 	bool connect_failed:1;      /* connection failed */
 	bool lost_prematurely:1;    /* lost connection before receiving any data */
@@ -272,11 +282,13 @@ struct http_client_peer {
 
 	/* active connections to this peer */
 	ARRAY_TYPE(http_client_connection) conns;
+	/* pending connections (not ready connecting) */
+	ARRAY_TYPE(http_client_connection) pending_conns;
 
 	/* zero time-out for consolidating request handling */
 	struct timeout *to_req_handling;
 
-	bool connecting:1;       /* peer is waiting to be connected */
+	bool connect_failed:1;   /* last connection attempt failed */
 	bool connect_backoff:1;  /* peer is waiting for backoff timout*/
 	bool disconnected:1;     /* peer is already disconnected */
 	bool handling_requests:1;/* currently running request handler */
@@ -344,6 +356,7 @@ struct http_client_host_shared {
 	/* timeouts */
 	struct timeout *to_idle;
 
+	bool destroyed:1;	/* shared host object is being destroyed */
 	bool unix_local:1;
 	bool explicit_ip:1;
 };
@@ -364,6 +377,8 @@ struct http_client {
 	struct http_client_context *cctx;
 	struct http_client_settings set;
 
+	struct http_client *prev, *next;
+
 	struct event *event;
 	struct ioloop *ioloop;
 	struct ssl_iostream_context *ssl_ctx;
@@ -385,9 +400,16 @@ struct http_client_context {
 	pool_t pool;
 	unsigned int refcount;
 	struct event *event;
+	struct ioloop *ioloop;
 
 	struct http_client_settings set;
 
+	struct dns_client *dns_client;
+	const char *dns_client_socket_path;
+	unsigned int dns_ttl_msecs;
+	unsigned int dns_lookup_timeout_msecs;
+
+	struct http_client *clients_list;
 	struct connection_list *conn_list;
 
 	HASH_TABLE_TYPE(http_client_peer_shared) peers;
@@ -468,16 +490,14 @@ void http_client_request_ref(struct http_client_request *req);
 bool http_client_request_unref(struct http_client_request **_req);
 void http_client_request_destroy(struct http_client_request **_req);
 
-int http_client_request_delay_from_response(struct http_client_request *req,
-	const struct http_response *response);
 void http_client_request_get_peer_addr(const struct http_client_request *req,
 	struct http_client_peer_addr *addr);
 enum http_response_payload_type
 http_client_request_get_payload_type(struct http_client_request *req);
 int http_client_request_send(struct http_client_request *req,
-			    bool pipelined, const char **error_r);
+			    bool pipelined);
 int http_client_request_send_more(struct http_client_request *req,
-				  bool pipelined, const char **error_r);
+				  bool pipelined);
 bool http_client_request_callback(struct http_client_request *req,
 	struct http_response *response);
 void http_client_request_connect_callback(struct http_client_request *req,
@@ -506,10 +526,15 @@ void http_client_connection_ref(struct http_client_connection *conn);
 bool http_client_connection_unref(struct http_client_connection **_conn);
 void http_client_connection_close(struct http_client_connection **_conn);
 
+void http_client_connection_lost(struct http_client_connection **_conn,
+				 const char *error) ATTR_NULL(2);
+
 void http_client_connection_peer_closed(struct http_client_connection **_conn);
 void http_client_connection_request_destroyed(
 	struct http_client_connection *conn, struct http_client_request *req);
 
+void http_client_connection_handle_output_error(
+	struct http_client_connection *conn);
 int http_client_connection_output(struct http_client_connection *conn);
 void http_client_connection_start_request_timeout(
 	struct http_client_connection *conn);
